
From: Jeff Dike <jdike@addtoit.com>

This makes SIGWINCH work again, and fixes a couple of SIGWINCH-associated
crashes.  First, the sigio thread disables SIGWINCH because all hell breaks
loose if it ever gets one and tries to call the signal handling code.  Second,
there was a problem with deferencing tty structs after they were freed.  The
SIGWINCH support for a tty wasn't being turned off or freed after the tty went
away.

Signed-off-by: Jeff Dike <jdike@addtoit.com>
Signed-off-by: Andrew Morton <akpm@osdl.org>
---

 arch/um/drivers/line.c        |   41 +++++++++++++++++++++++++++++++++++++++--
 arch/um/include/os.h          |    1 +
 arch/um/kernel/irq.c          |    1 -
 arch/um/kernel/process.c      |    2 --
 arch/um/kernel/sigio_user.c   |    1 +
 arch/um/kernel/skas/process.c |    7 ++++++-
 arch/um/kernel/tt/tracer.c    |    4 +++-
 arch/um/os-Linux/process.c    |    5 +++++
 8 files changed, 55 insertions(+), 7 deletions(-)

diff -puN arch/um/drivers/line.c~uml-fix-sigwinch-relaying arch/um/drivers/line.c
--- 25/arch/um/drivers/line.c~uml-fix-sigwinch-relaying	2005-05-02 19:01:43.000000000 -0700
+++ 25-akpm/arch/um/drivers/line.c	2005-05-02 19:01:43.000000000 -0700
@@ -462,12 +462,15 @@ out:
 	return err;
 }
 
+static void unregister_winch(struct tty_struct *tty);
+
 void line_close(struct tty_struct *tty, struct file * filp)
 {
 	struct line *line = tty->driver_data;
 
-	/* XXX: I assume this should be called in process context, not with interrupt
-	 * disabled!*/
+	/* XXX: I assume this should be called in process context, not with
+         *  interrupts disabled!
+         */
 	spin_lock_irq(&line->lock);
 
 	/* We ignore the error anyway! */
@@ -478,6 +481,12 @@ void line_close(struct tty_struct *tty, 
 		line_disable(tty, -1);
 		tty->driver_data = NULL;
 	}
+
+        if((line->count == 0) && line->sigio){
+                unregister_winch(tty);
+                line->sigio = 0;
+        }
+
 	spin_unlock_irq(&line->lock);
 }
 
@@ -729,6 +738,34 @@ void register_winch_irq(int fd, int tty_
 	up(&winch_handler_sem);
 }
 
+static void unregister_winch(struct tty_struct *tty)
+{
+	struct list_head *ele;
+	struct winch *winch, *found = NULL;
+
+	down(&winch_handler_sem);
+	list_for_each(ele, &winch_handlers){
+		winch = list_entry(ele, struct winch, list);
+                if(winch->tty == tty){
+                        found = winch;
+                        break;
+                }
+        }
+
+        if(found == NULL)
+                goto out;
+
+        if(winch->pid != -1)
+                os_kill_process(winch->pid, 1);
+
+        free_irq_by_irq_and_dev(WINCH_IRQ, winch);
+        free_irq(WINCH_IRQ, winch);
+        list_del(&winch->list);
+        kfree(winch);
+ out:
+	up(&winch_handler_sem);
+}
+
 static void winch_cleanup(void)
 {
 	struct list_head *ele;
diff -puN arch/um/include/os.h~uml-fix-sigwinch-relaying arch/um/include/os.h
--- 25/arch/um/include/os.h~uml-fix-sigwinch-relaying	2005-05-02 19:01:43.000000000 -0700
+++ 25-akpm/arch/um/include/os.h	2005-05-02 19:01:43.000000000 -0700
@@ -160,6 +160,7 @@ extern void os_kill_process(int pid, int
 extern void os_kill_ptraced_process(int pid, int reap_child);
 extern void os_usr1_process(int pid);
 extern int os_getpid(void);
+extern int os_getpgrp(void);
 
 extern int os_map_memory(void *virt, int fd, unsigned long long off,
 			 unsigned long len, int r, int w, int x);
diff -puN arch/um/kernel/irq.c~uml-fix-sigwinch-relaying arch/um/kernel/irq.c
--- 25/arch/um/kernel/irq.c~uml-fix-sigwinch-relaying	2005-05-02 19:01:43.000000000 -0700
+++ 25-akpm/arch/um/kernel/irq.c	2005-05-02 19:01:43.000000000 -0700
@@ -163,7 +163,6 @@ void __init init_IRQ(void)
 		irq_desc[i].handler = &SIGIO_irq_type;
 		enable_irq(i);
 	}
-	init_irq_signals(0);
 }
 
 /*
diff -puN arch/um/kernel/process.c~uml-fix-sigwinch-relaying arch/um/kernel/process.c
--- 25/arch/um/kernel/process.c~uml-fix-sigwinch-relaying	2005-05-02 19:01:43.000000000 -0700
+++ 25-akpm/arch/um/kernel/process.c	2005-05-02 19:01:43.000000000 -0700
@@ -65,8 +65,6 @@ void init_new_thread_signals(int altstac
 		    SIGUSR1, SIGIO, SIGWINCH, SIGALRM, SIGVTALRM, -1);
 	set_handler(SIGBUS, (__sighandler_t) sig_handler, flags, 
 		    SIGUSR1, SIGIO, SIGWINCH, SIGALRM, SIGVTALRM, -1);
-	set_handler(SIGWINCH, (__sighandler_t) sig_handler, flags, 
-		    SIGUSR1, SIGIO, SIGWINCH, SIGALRM, SIGVTALRM, -1);
 	set_handler(SIGUSR2, (__sighandler_t) sig_handler, 
 		    flags, SIGUSR1, SIGIO, SIGWINCH, SIGALRM, SIGVTALRM, -1);
 	signal(SIGHUP, SIG_IGN);
diff -puN arch/um/kernel/sigio_user.c~uml-fix-sigwinch-relaying arch/um/kernel/sigio_user.c
--- 25/arch/um/kernel/sigio_user.c~uml-fix-sigwinch-relaying	2005-05-02 19:01:43.000000000 -0700
+++ 25-akpm/arch/um/kernel/sigio_user.c	2005-05-02 19:01:43.000000000 -0700
@@ -182,6 +182,7 @@ static int write_sigio_thread(void *unus
 	int i, n, respond_fd;
 	char c;
 
+        signal(SIGWINCH, SIG_IGN);
 	fds = &current_poll;
 	while(1){
 		n = poll(fds->poll, fds->used, -1);
diff -puN arch/um/kernel/skas/process.c~uml-fix-sigwinch-relaying arch/um/kernel/skas/process.c
--- 25/arch/um/kernel/skas/process.c~uml-fix-sigwinch-relaying	2005-05-02 19:01:43.000000000 -0700
+++ 25-akpm/arch/um/kernel/skas/process.c	2005-05-02 19:01:43.000000000 -0700
@@ -28,10 +28,11 @@
 #include "chan_user.h"
 #include "signal_user.h"
 #include "registers.h"
+#include "process.h"
 
 int is_skas_winch(int pid, int fd, void *data)
 {
-	if(pid != os_getpid())
+        if(pid != os_getpgrp())
 		return(0);
 
 	register_winch_irq(-1, fd, -1, data);
@@ -259,6 +260,10 @@ int start_idle_thread(void *stack, void 
 	sigjmp_buf **switch_buf = switch_buf_ptr;
 	int n;
 
+	set_handler(SIGWINCH, (__sighandler_t) sig_handler,
+		    SA_ONSTACK | SA_RESTART, SIGUSR1, SIGIO, SIGALRM,
+		    SIGVTALRM, -1);
+
 	*fork_buf_ptr = &initial_jmpbuf;
 	n = sigsetjmp(initial_jmpbuf, 1);
 	if(n == 0)
diff -puN arch/um/kernel/tt/tracer.c~uml-fix-sigwinch-relaying arch/um/kernel/tt/tracer.c
--- 25/arch/um/kernel/tt/tracer.c~uml-fix-sigwinch-relaying	2005-05-02 19:01:43.000000000 -0700
+++ 25-akpm/arch/um/kernel/tt/tracer.c	2005-05-02 19:01:43.000000000 -0700
@@ -26,6 +26,7 @@
 #include "kern_util.h"
 #include "chan_user.h"
 #include "ptrace_user.h"
+#include "irq_user.h"
 #include "mode.h"
 #include "tt.h"
 
@@ -33,7 +34,7 @@ static int tracer_winch[2];
 
 int is_tracer_winch(int pid, int fd, void *data)
 {
-	if(pid != tracing_pid)
+	if(pid != os_getpgrp())
 		return(0);
 
 	register_winch_irq(tracer_winch[0], fd, -1, data);
@@ -119,6 +120,7 @@ static int signal_tramp(void *arg)
 	signal(SIGSEGV, (__sighandler_t) sig_handler);
 	set_cmdline("(idle thread)");
 	set_init_pid(os_getpid());
+	init_irq_signals(0);
 	proc = arg;
 	return((*proc)(NULL));
 }
diff -puN arch/um/os-Linux/process.c~uml-fix-sigwinch-relaying arch/um/os-Linux/process.c
--- 25/arch/um/os-Linux/process.c~uml-fix-sigwinch-relaying	2005-05-02 19:01:43.000000000 -0700
+++ 25-akpm/arch/um/os-Linux/process.c	2005-05-02 19:01:43.000000000 -0700
@@ -123,6 +123,11 @@ int os_getpid(void)
 	return(getpid());
 }
 
+int os_getpgrp(void)
+{
+	return getpgrp();
+}
+
 int os_map_memory(void *virt, int fd, unsigned long long off, unsigned long len,
 		  int r, int w, int x)
 {
_
